;;;; egg:       disjoint-sets
;;;; file:      disjoint-sets.scm
;;;; author:    elf <elf@ephemeral.net>
;;;; date:      02 MAy 2008
;;;; licence:   BSD (see LICENCE)
;;;; version:   1.0
;;;; purpose:   procedures for operating on disjoint sets
;;;;
;;;; history:   1.0  20080502 (elf)  Initial release
;;;;




;;; chicken library loading


(use srfi-23)   ; error reporting



;;; chicken compile-time directives


(eval-when (compile)
    (declare
        (uses library)
        (bound-to-procedure
            check-dset
            check-eint
            %disjoint-set-ref
            %disjoint-set-update-path
            %disjoint-set-clean-all
            make-disjoint-set
            disjoint-set?
            disjoint-set-ref
            disjoint-set-unify!
            disjoint-set-merge!
            disjoint-set-equal?
            disjoint-set-members
            disjoint-set-elements
            disjoint-set-sets
            disjoint-set-unified?
            disjoint-set->list
            )
        (export
            make-disjoint-set
            disjoint-set?
            disjoint-set-ref
            disjoint-set-unify!
            disjoint-set-merge!
            disjoint-set-equal?
            disjoint-set-members
            disjoint-set-elements
            disjoint-set-sets
            disjoint-set-unified?
            disjoint-set->list
            )
        (emit-exports "disjoint-sets.exports")
        (inline)
        (inline-limit 100)
        (lambda-lift)
        (disable-interrupts)
        (fixnum-arithmetic)
        (no-bound-checks)
        (no-procedure-checks)
        (standard-bindings)
        (extended-bindings)
        (usual-integrations)
    ))



;;; disjoint set structure:
;;;
;;; 0: tag
;;; 1: number of elements
;;; 2: number of merges
;;; 3: vector of elements



;;; typechecking and other auxiliary procedures


;; (check-dset OBJ PNAME)
;; signal an error if OBJ is not a disjoint set
(define-inline (check-dset obj pname)
    (or (##sys#structure? obj 'disjoint-set)
        (##sys#signal-hook #:type-error pname
                           "bad argument type - not a disjoint set"
                           obj)))

;; (check-eint OBJ DSET PNAME)
;; signal an error if OBJ is not an exact integer and a valid index for DSET
(define-inline (check-eint obj dset pname)
    (or (and (integer? obj) (exact? obj) (> obj -1))
        (##sys#signal-hook #:type-error pname
                           "bad argument type - not a integer index value"
                           obj))
    (or (< obj (##sys#slot dset 1))
        (##sys#signal-hook #:bounds-error pname
                           "out of range"
                           obj (##sys#slot dset 1))))



;;; creation and query procedures


;; (make-disjoint-set NUM-ELEMENTS)
;; create a new disjoint set
(define (make-disjoint-set nelem)
    (or (and (integer? nelem) (exact? nelem) (> nelem 0))
        (##sys#signal-hook #:type-error 'make-disjoint-set
                           "bad argument type - not an exact positive integer"
                           nelem))
    (##sys#make-structure 'disjoint-set nelem 0 (make-vector nelem #f)))

;; (disjoint-set? OBJ)
;; type predicate for disjoint sets
(define (disjoint-set? obj)
    (##sys#structure? obj 'disjoint-set))


;; (%disjoint-set-ref DISJOINT-SET-VEC ELEM)
;; return representative element, not typechecked
(define-inline (%disjoint-set-ref dsetvec elem)
    (let loop ((i   elem)
               (n   (vector-ref dsetvec elem)))
        (if n
            (loop n (vector-ref dsetvec n))
            i)))

;; (disjoint-set-ref DISJOINT-SET ELEM)
;; return the representative element for the set containing ELEM
(define (disjoint-set-ref dset elem)
    (check-dset dset 'disjoint-set-ref)
    (check-eint elem dset 'disjoint-set-ref)
    (%disjoint-set-ref (##sys#slot dset 3) elem))

;; (disjoint-set-elements DISJOINT-SET)
;; return the number of elements of DISJOINT-SET
(define (disjoint-set-elements dset)
    (check-dset dset 'disjoint-set-elements)
    (##sys#slot dset 1))

;; (disjoint-set-sets DISJOINT-SET)
;; return the number of unique sets 
(define (disjoint-set-sets dset)
    (check-dset dset 'disjoint-set-sets)
    (- (##sys#slot dset 1) (##sys#slot dset 2)))

;; (disjoint-set-unified? DISJOINT-SET)
;; return #t if there is only one set
(define (disjoint-set-unified? dset)
    (check-dset dset 'disjoint-set-unified?)
    (= 1 (- (##sys#slot dset 1) (##sys#slot dset 2))))

;; (disjoint-set-equal? DISJOINT-SET ELEM1 ELEM2)
;; return #t if ELEM1 and ELEM2 are members of the same set
(define (disjoint-set-equal? dset elem1 elem2)
    (check-dset dset 'disjoint-set-equal?)
    (check-eint elem1 dset 'disjoint-set-equal?)
    (check-eint elem2 dset 'disjoint-set-equal?)
    (= (%disjoint-set-ref (##sys#slot dset 3) elem1)
       (%disjoint-set-ref (##sys#slot dset 3) elem2)))



;;; unification


;; (%disjoint-set-update-path DISJOINT-SET-VEC ELEM RVAL)
;; perform path compression on a lookup path
(define-inline (%disjoint-set-update-path dsetvec elem rval)
    (or (= elem rval)
        (let loop ((i   elem)
                   (n   (vector-ref dsetvec elem)))
            (vector-set! dsetvec i rval)
            (or (= n rval)
                (loop n (vector-ref dsetvec n))))))

;; (disjoint-set-unify! DISJOINT-SET ELEM1 ELEM2)
;; unify the sets containing ELEM1 and ELEM2
(define (disjoint-set-unify! dset elem1 elem2)
    (check-dset dset 'disjoint-set-unify!)
    (check-eint elem1 dset 'disjoint-set-unify!)
    (check-eint elem2 dset 'disjoint-set-unify!)
    (let ((r1   (%disjoint-set-ref (##sys#slot dset 3) elem1))
          (r2   (%disjoint-set-ref (##sys#slot dset 3) elem2)))
        (and (= r1 r2)
             (##sys#signal-hook #:bounds-error 'disjoint-set-unify!
                                "disjoint sets already unified"
                                elem1 elem2 r1 r2))
        (let ((m   (min r1 r2))
              (x   (max r1 r2)))
            (vector-set! (##sys#slot dset 3) x m)
            (##sys#setslot dset 2 (+ 1 (##sys#slot dset 2)))
            (%disjoint-set-update-path (##sys#slot dset 3) elem1 m)
            (%disjoint-set-update-path (##sys#slot dset 3) elem2 m))))

;; (disjoint-set-merge! DISJOINT-SET ELEM1 ELEM2)
;; another name for disjoint-set-unify!
(define disjoint-set-merge! disjoint-set-unify!)



;;; debugging and export


;; (%disjoint-set-iota NUM-ELEMS)
;; reverse iota
;; (define-inline (%disjoint-set-iota nelem)
;;    (let loop ((i   0)
;;               (r   '()))
;;        (if (= i nelem)
;;            r
;;            (loop (+ 1 i) (cons i r)))))

;; (%disjoint-set-clean-all DISJOINT-SET-VEC NUM-ELEMS)
;; perform path compression on everything
(define-inline (%disjoint-set-clean-all dsetvec nelem)
    (let loop ((x   (- nelem 1))
               (i   (- nelem 1))
               (n   (vector-ref dsetvec (- nelem 1)))
               (r   '()))
        (if n
            (loop x n (vector-ref dsetvec n) (cons i r))
            (let ((x   (- x 1)))
                (for-each
                    (cute vector-set! dsetvec <> i)
                    r)
                (if (< x 0)
                    #t
                    (loop x x (vector-ref dsetvec x) '()))))))

;; (disjoint-set-members DISJOINT-SET ELEM)
;; return all members of the set containing ELEM as a list
(define (disjoint-set-members dset elem)
    (check-dset dset 'disjoint-set-members)
    (check-eint elem dset 'disjoint-set-members)
    (%disjoint-set-clean-all (##sys#slot dset 3) (##sys#slot dset 1))
    (let loop ((l   (- (##sys#slot dset 1) 1))
               (i   (or (vector-ref (##sys#slot dset 3) elem) elem))
               (r   '()))
        (if (< l 0)
            r
            (if (eq? i (vector-ref (##sys#slot dset 3) l))
                (loop (- l 1) i (cons l r))
                (loop (- l 1) i r)))))

;; (disjoint-set->list DISJOINT-SET)
;; convert a disjoint set into a list of lists
(define (disjoint-set->list dset)
    (check-dset dset 'disjoint-set->list)
    (%disjoint-set-clean-all (##sys#slot dset 3) (##sys#slot dset 1))
    (let loop ((l   (- (##sys#slot dset 1) 1))
               (v   (make-vector (##sys#slot dset 1) '())))
        (if (< l 0)
            (let loop2 ((l   (- (##sys#slot dset 1) 1))
                        (r   '()))
                (if (< l 0)
                    r
                    (if (null? (vector-ref v l))
                        (loop2 (- l 1) r)
                        (loop2 (- l 1) (cons (vector-ref v l) r)))))
            (let ((t   (or (vector-ref (##sys#slot dset 3) l) l)))
                (vector-set! v t (cons l (vector-ref v t)))
                (loop (- l 1) v)))))

